/**
 * NORRIS-IoC.js
 * Copyright (c) 2011 Ruben LZ Tan
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
 * IN THE SOFTWARE.
 */
'use strict';

///////////////////////////////////////////////////////////////////////////////
// @nordoc[package] ioc
//
// version
// -    refactored code to use the standard NORRIS creational pattern
// -    hoisted all variables
// -    changed the event 'error' to 'failure'
// -    added documentation
///////////////////////////////////////////////////////////////////////////////
module.exports.make = function make(deps) {
    if (undefined === deps) deps = {};

    ///////////////////////////////////////////////////////////////////////////
    // Hoisted variables
    ///////////////////////////////////////////////////////////////////////////
    var base, npm, fs, Emitter, installer, log,
        loadNpm, loadNode, handleNpm, handleNode, _ito, IoC;

    ///////////////////////////////////////////////////////////////////////////
    // Dependency injection/definition
    ///////////////////////////////////////////////////////////////////////////
    base        = deps.base         || process.cwd(),
    npm         = deps.npm          || require('npm'),
    fs          = deps.fs           || require('fs'),
    Emitter     = deps.Emitter      || require('events').EventEmitter,
    installer   = deps.installer    || require(__dirname + '/installer').make({ npm : npm, fs : fs, base : base }),

    loadNpm     = deps.loadNpm      || require(__dirname + '/loader/npm').make({ installer : installer }),
    loadNode    = deps.loadNode     || require(__dirname + '/loader/node').make(),

    handleNpm   = deps.handleNpm    || require(__dirname + '/handle/npm').make(),
    handleNode  = deps.handleNode   || require(__dirname + '/handle/node').make()
    ;

    //-------------------------------------------------------------------------
    // @nordoc[function]
    //
    // @desc    Works similarly to Array.forEach(), calling the callback on
    //          every object property.
    //
    // @args    o [object]
    //          The object to be passed in.
    //
    //          cb [function] (key, value)
    //          Called on every iteration, passing in the key and value as
    //          arguments to the callback
    //
    // @return  none
    //-------------------------------------------------------------------------
    _ito = function _ito(o, cb) {
        var o, i;

        for (i in o) {
            cb(i, o[i]);
        }
    }

    ///////////////////////////////////////////////////////////////////////////
    // @nordoc[module] ioc
    ///////////////////////////////////////////////////////////////////////////
    IoC = {
        //---------------------------------------------------------------------
        // @nordoc[property]
        //
        // @desc    The base path/current working directory of the application
        //---------------------------------------------------------------------
        basepath : base,

        //---------------------------------------------------------------------
        // @nordoc[property]
        //
        // @desc    Container of loader functions for the various registered
        //          signatures. The loader function decides on how to load a
        //          particular signature.
        //---------------------------------------------------------------------
        loader : {
            npm     : loadNpm, 
            node    : loadNode
        },

        //---------------------------------------------------------------------
        // @nordoc[property]
        //
        // @desc    Container of handler functions for the various registered
        //          signatures. The handler function is called AFTER a module
        //          is loaded, usually used to decide where to save the loaded
        //          module to (in relation to the deps object, usually)
        //---------------------------------------------------------------------
        handler : {
            npm     : handleNpm,
            node    : handleNode,
        },

        //---------------------------------------------------------------------
        // @nordoc[property]
        //
        // @desc    Dependencies currently loaded
        //---------------------------------------------------------------------
        deps : {},

        //---------------------------------------------------------------------
        // @nordoc[property]
        //
        // @desc    Logs
        //---------------------------------------------------------------------
        log : [],

        //---------------------------------------------------------------------
        // @nordoc[method]
        //
        // @desc    Registers a new signature loader
        //
        // @args    signature [string]
        //          The name of the signature to be registered
        //
        //          cb [function] (module, callback)
        //          The callback to be called when encountering this specific
        //          load signature
        //
        // @return  Returns itself so that other methods can be chained
        //---------------------------------------------------------------------
        registerLoader : function registerLoader(signature, cb) {
            this.loader[signature] = cb;
            return this;
        },

        //---------------------------------------------------------------------
        // @nordoc[method]
        //
        // @desc    Registers a new signature handler
        //
        // @args    signature [string]
        //          The name of the signature to be registered
        //
        //          cb [function] 
        //          The callback to be called when encountering this specific
        //          handler signature
        //
        // @return  Returns itself so that other methods can be chained
        //---------------------------------------------------------------------
        registerHandler : function registerHandler(signature, cb) {
            this.handler[signature] = cb;
            return this;
        },

        //---------------------------------------------------------------------
        // @nordoc[method]
        //
        // @desc    Adds a dependency object into the loading list
        //
        // @args    deps [object]
        //          A dependency object containing load signatures for IoC to
        //          load.
        //
        // @return  Returns itself so that this operation can be chained
        //---------------------------------------------------------------------
        load : function load(deps) {
            var deps, key, dep, loadList;

            loadList = this.deps;

            for (key in deps) {

                // We reverse the key and value here because signatures are
                // unique, while keys aren't. In a typical signature like this:
                // `fs : '[npm] file'`, `fs` is the "key", while '[npm] file'
                // is the "signature". The signature tells the loader how to
                // load the file, while the key tells the handler what is the
                // name of the property that it is to save the loaded module
                // to upon success. This way, the handler gets more freedom
                // to decide how to handle (where to save) the loaded
                // dependency.
                dep = deps[key];
                loadList[dep] = key;
            }

            return this;
        },

        //---------------------------------------------------------------------
        // @nordoc[method]
        //
        // @desc    Starts the loading process. Upon completion, it will call
        //          the callback passed in as an argument.
        //
        // @args    cb [function] (deps)
        //          The callback to call upon completion
        //
        // @return  Returns a promise that emits the event 'success' upon
        //          successful completion of loading, or 'failure' upon
        //          any failures. The event 'error' is emitted on any
        //          exceptions that are thrown.
        //---------------------------------------------------------------------
        ready : function ready(cb) {
            var self, overall, loaderTypes, handlerTypes, results, npmConfig, 
                base, deps, _ragequit, _postProcessDeps;

            cb              = (cb) ? cb : function() {},    // callback to call
            self            = this,                         // self reference
            base            = self.basepath,                // the base dir
            overall         = new Emitter(),                // overall progress
            loaderTypes     = Object.keys(self.loader),     // all loader types
            handlerTypes    = Object.keys(self.handler),    // all handler types
            results         = {},                           // stores loaded deps
            npmConfig       = {
                loglevel : 'silent',
                outfd    : ''
            },
            deps            = self.deps;

            // Function to rage quit (on critical errors)
            _ragequit = function _ragequit(key, err) {
                if (null !== key) delete deps[key];

                if (Object.keys(deps).length <= 0 || null === key) {
                    process.nextTick(function onTick() {
                        cb(err);
                        overall.emit('failure', err);
                    });
                }
            }

            // Function to post process the dependencies
            _postProcessDeps = function _postProcessDeps(key, results) {
                if (null !== key) delete deps[key];

                if (Object.keys(deps).length <= 0 || null === key) {
                    process.nextTick(function onTick() {
                        cb(null, results);
                        overall.emit('success', results);
                    });
                }
            }

            // If no dependencies is loaded, return immediately
            if (0 === Object.keys(deps).length) {
                _ragequit(null, new Error('No dependencies to load. Call load() before ready()'));
                return;
            }

            // Start loading!
            npm.load(npmConfig, function onNpmLoad() {
                // Parse each signature and load it accordingly
                _ito(deps, function onEach(signature, key) {
                    var start, end, type, module;

                    start   = signature.indexOf('[');
                    end     = signature.indexOf(']');
                    type    = signature.slice(start + 1, end);
                    module  = signature.slice(end + 1).trim();

                    // If there's an error in parsing this dependency, remove it
                    // and continue
                    if (!type || !module || !key) {
                        _postProcessDeps(signature, results);
                        return;
                    }

                    if (undefined === self.loader[type] || undefined === self.handler[type]) {
                        _ragequit(signature, new Error('Loader and/or handler functions for the signature type [' + type + '] not found'));
                        return;
                    }

                    // At this point we have all the information we need to start
                    // the loading process. "type" tells us which loader/handler
                    // signature to use, "module" tells the loader what to load,
                    // while "key" tells the handler where to save the loaded
                    // dependency to. Life is good.
                    self.loader[type].call(self, module, function onLoad(err, mod) {

                        // Do not proceed if loading has failed
                        if (err) {
                            _ragequit(signature, err);
                            return;
                        }

                        self.handler[type].call(self, key, mod, results, function onHandle(err, results) {
                            if (err) {
                                _ragequit(signature, err);
                                return;
                            }

                            _postProcessDeps(signature, results);
                        });
                    });
                });
            });

            return overall;
        }
    }

    return IoC;
}